//
// CheesyPlayer
// by Eric Shepherd
//
// Copyright 1999, Be Incorporated.
//
// Module:	MediaPlayer.cpp
//			Player class code.
//

#include "MediaPlayer.h"
#include "MediaApp.h"

#include <MediaRoster.h>
#include <MediaAddOn.h>
#include <MediaDefs.h>
#include <MediaNode.h>
#include <TimeSource.h>
#include <Errors.h>

#include <Path.h>
#include <Entry.h>
#include <stdio.h>
#include <string.h>

//
// MediaPlayer::MediaPlayer
//
// Construct the player object.
//
MediaPlayer::MediaPlayer(entry_ref *ref) {
	dormant_node_info nodeInfo;		// Info about dormant nodes
	status_t err;
	
	playingFlag = false;
	roster = BMediaRoster::Roster();		// Get the roster object
	
	// Find a node that can handle the file.
	
	initStatus = roster->SniffRef(*ref, 0, &nodeInfo);
	if (initStatus) {
		return;
	}
	
	// Instantiate a copy of the node for our use.
	
	initStatus = roster->InstantiateDormantNode(nodeInfo, &mediaFileNode);
	if (initStatus) {
		return;
	}
	
	// Tell the file handling node what file to handle.
	
	roster->SetRefFor(mediaFileNode, *ref, false, &duration);
	if ((err = Setup()) != B_OK) {
		printf("Error %08lX in Setup()\n", err);
		initStatus = err;
	}
	else {
		initStatus = Start();
	}
}

//
// MediaPlayer::~MediaPlayer
//
// Unhook from the media server and do other cleanup stuff.
//
MediaPlayer::~MediaPlayer() {
	status_t err;
	
	// Don't disconnect unless stopped.
	
	Stop();
	
	// Now we can disconnect the nodes
	
	err = roster->Disconnect(mediaFileNode.node, fileNodeOutput.source,
					 codecNode.node, codecInput.destination);
	if (err != B_OK) {
		printf("Error disconnecting file from codec %08lX\n", err);
	}
	err = roster->Disconnect(codecNode.node, codecOutput.source,
					 videoNode.node, videoInput.destination);
	if (err != B_OK) {
		printf("Error disconnecting codec from video %08lX\n", err);
	}
	err = roster->Disconnect(mediaFileNode.node, fileAudioOutput.source,
					 audioNode.node, audioInput.destination);
	if (err != B_OK) {
		printf("Error disconnecting file from audio %08lX\n", err);
	}
	
	// Release nodes
	
	roster->ReleaseNode(codecNode);
	roster->ReleaseNode(mediaFileNode);
}

//
// MediaPlayer::InitCheck
//
// Returns the error code resulting from the constructor.
// If this returns anything but B_OK, don't use the object!
//
status_t MediaPlayer::InitCheck(void) {
	return initStatus;
}

//
// MediaPlayer::Setup
//
// Prepare the file and video window for playback.
//
#define RETERR(x) if ((x)!=B_OK){return (x);}
status_t MediaPlayer::Setup(void) {
	status_t err;
	media_format tryFormat;
	dormant_node_info nodeInfo;
	int32 nodeCount;

	// Get a refernece to the default audio mixer
	
	err = roster->GetAudioMixer(&audioNode);
	RETERR(err);

	// Get a reference to the default video output node
	
	err = roster->GetVideoOutput(&videoNode);	// Get reference to an appropriate video out node
	RETERR(err);
	
	// Get a reference to the time source.
	
	err = roster->GetTimeSource(&timeSourceNode);
	RETERR(err);
	b_timesource = roster->MakeTimeSourceFor(timeSourceNode);

	// Get an audio output from the file handler node.
	
	err = roster->GetFreeOutputsFor(mediaFileNode, &fileAudioOutput, 1,
					&fileAudioCount, B_MEDIA_RAW_AUDIO);
	RETERR(err);

	// Get an equal number of inputs into the audio mixer node.
	
	err = roster->GetFreeInputsFor(audioNode, &audioInput, fileAudioCount,
					&audioInputCount, B_MEDIA_RAW_AUDIO);
	RETERR(err);

	// Get a video output from the file handler node.

	err = roster->GetFreeOutputsFor(mediaFileNode, &fileNodeOutput, 1,
					&fileOutputCount, B_MEDIA_ENCODED_VIDEO);
	RETERR(err);

	// Get an equal number of inputs into the video output node.
	
	err = roster->GetFreeInputsFor(videoNode, &videoInput, fileOutputCount,
					&videoInputCount, B_MEDIA_RAW_VIDEO);
	RETERR(err);

	// Set up the video tracks by connecting them to the codec, then
	// from there to the video output node.  First, get info about
	//the node needed to handle the media file's format.
	
	nodeCount = 1;
	err = roster->GetDormantNodes(&nodeInfo, &nodeCount,
					&fileNodeOutput.format);
	RETERR(err);
	if (!nodeCount) {
		printf("Can't find the needed codec.\n");
		return -1;
	}
	
	// Instantiate the appropriate codec.
	
	err = roster->InstantiateDormantNode(nodeInfo, &codecNode);
	RETERR(err);
	
	// Find inputs into the node..
	
	err = roster->GetFreeInputsFor(codecNode, &codecInput, 1, &nodeCount, 
					B_MEDIA_ENCODED_VIDEO);
	RETERR(err);

	// Find outputs from the node.

	err = roster->GetFreeOutputsFor(codecNode, &codecOutput, 1, &nodeCount, 
					B_MEDIA_RAW_VIDEO);
	RETERR(err);
	
	// This is the format the codec needs to accept as input; we peel it
	// right out of the media file output's information.
	
	tryFormat = fileNodeOutput.format;
	
	// Make the connection between the file reader node and the codec.
	
	err = roster->Connect(fileNodeOutput.source, codecInput.destination,
					&tryFormat, &fileNodeOutput, &codecInput);
	RETERR(err);
	
	// Set up a format to link the codec to the video output node.
	
	tryFormat.type = B_MEDIA_RAW_VIDEO;
	tryFormat.u.raw_video = media_raw_video_format::wildcard;
	tryFormat.u.raw_video.display.line_width = codecInput.format.u.encoded_video.output.display.line_width;
	tryFormat.u.raw_video.display.line_count = codecInput.format.u.encoded_video.output.display.line_count;

	// Connect the codec to the video node.

	err = roster->Connect(codecOutput.source, videoInput.destination,
					&tryFormat, &codecOutput, &videoInput);
	RETERR(err);
	
	// Connect the media file's audio to the audio mixer.
	
	tryFormat = fileAudioOutput.format;
	err = roster->Connect(fileAudioOutput.source, audioInput.destination,
					&tryFormat, &fileAudioOutput, &audioInput);
	RETERR(err);

	// Set all the nodes' time sources.

	err = roster->SetTimeSourceFor(mediaFileNode.node, timeSourceNode.node);
	RETERR(err);
	err = roster->SetTimeSourceFor(videoNode.node, timeSourceNode.node);
	RETERR(err);
	err = roster->SetTimeSourceFor(codecOutput.node.node, timeSourceNode.node);
	RETERR(err);
	err = roster->SetTimeSourceFor(audioNode.node, timeSourceNode.node);
	RETERR(err);
	return B_OK;
}

//
// MediaPlayer::Start
//
// Start playing the movie.
//
status_t MediaPlayer::Start(void) {
	status_t err;
	err = roster->GetStartLatencyFor(timeSourceNode, &startTime);
	if (err != B_OK) {
		return err;
	}
	startTime += b_timesource->PerformanceTimeFor(BTimeSource::RealTime()
					+ 1000000 / 50);
	
	err = roster->StartNode(mediaFileNode, startTime);
	if (err != B_OK) {
		return err;
	}
	err = roster->StartNode(codecNode, startTime);
	if (err != B_OK) {
		return err;
	}
	err = roster->StartNode(videoNode, startTime);
	if (err != B_OK) {
		return err;
	}
	playingFlag = true;
		
	return B_OK;
}

//
// MediaPlayer::Stop
//
// Start playing the movie.
//
status_t MediaPlayer::Stop(void) {
	status_t err;
	
	err = roster->StopNode(mediaFileNode, 0, true);
	if (err != B_OK) {
		return err;
	}
	err = roster->StopNode(codecNode, 0, true);
	if (err != B_OK) {
		return err;
	}
	err = roster->StopNode(videoNode, 0, true);
	if (err != B_OK) {
		return err;
	}

	playingFlag = false;
	return B_OK;
}

//
// Playing
//
// Is it still playing?
//
bool MediaPlayer::Playing(void) {
	bigtime_t currentTime;
	
	// If we know it's not playing, just say so.
	
	if (!playingFlag) {
		return false;
	}
	
	// Are we at or beyond the end of the movie?
	
	currentTime = b_timesource->PerformanceTimeFor(BTimeSource::RealTime());
	if (currentTime >= startTime+duration) {
		playingFlag = false;
	}
	return playingFlag;
}